/*	$Id: icetools.c,v 1.4 2012/03/28 15:03:33 kristaps Exp $ */
/*
 * Copyright (c) 2012 Kristaps Dzonsons <kristaps@bsd.lv>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#include <sys/param.h>
#include <sys/wait.h>

#include <assert.h>
#include <errno.h>
#include <expat.h>
#include <poll.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "extern.h"

#define	SERVER "http://dir.xiph.org/yp.xml"

/*
 * Use a simple state machine to keep track of where we are in the
 * parsed document.
 * Also use this to make sure the directory file is intact.
 */
enum	state {
	STATE_DIR = 0,
	STATE_ENTRY,
	STATE_NAME,
	STATE_URL,
	STATE_GENRE,
	STATE_SONG,
	STATE__MAX
};

/*
 * Keep track of some state bits in our "struct stream" counterpart:
 * sizes of read data and stuff we're going to convert.
 */
struct	entry {
	char		 *song;
	size_t		  songsz;
	size_t		  namesz;
	char		 *url;
	size_t		  urlsz;
	char		 *genre;
	size_t		  genresz;
};

/*
 * Keep track of an in-progress parse of a YP.
 * This consists of the "stream" and "genre" tables we're going to copy
 * into our output "struct icecast", alongwith the state bits we
 * maintain while building these structures.
 */
struct	ypdir {
	struct entry	 *e;
	struct stream	 *s;
	size_t		  ssz;
	size_t		  max;
	enum state	  state;
	char 		**g;
	size_t		 *gcount;
	size_t		  gsz;
};

/*
 * Used to control the state machine and underlying XML parse.
 */
struct	callback {
	void		(*start)(struct ypdir *);
	void		(*end)(struct ypdir *);
	const char	 *name;
	enum state	  enter;
};

static	void	 dupes(struct ypdir *);
static	void	 entry(struct ypdir *);
static	void	 entry_free(struct entry *);
static	void	 genre(struct ypdir *);
static	int	 state(struct ypdir *, enum state, enum state);
static	void	 stream_free(struct stream *);
static	void	 xml_char(void *, const XML_Char *, int);
static	void	 xml_elem_start(void *, 
			const XML_Char *, const XML_Char **);
static	void	 xml_elem_end(void *, const XML_Char *);

static	const struct callback states[STATE__MAX] = {
	{ NULL, NULL, "directory", STATE__MAX }, /* STATE_DIR */
	{ entry, dupes, "entry", STATE_DIR }, /* STATE_ENTRY */
	{ NULL, NULL, "server_name", STATE_ENTRY }, /* STATE_NAME */
	{ NULL, NULL, "listen_url", STATE_ENTRY }, /* STATE_URL */
	{ NULL, genre, "genre", STATE_ENTRY }, /* STATE_GENRE */
	{ NULL, NULL, "current_song", STATE_ENTRY }, /* STATE_SONG */
};

/*
 * A "soft" parse error, i.e., one not affecting the XML integrity or
 * the actual read process.
 * Means that the document contents are not a YP.
 */
static	int	error = 0;

void
icetools_free(struct icetools *p)
{
	size_t		 i;

	if (NULL == p)
		return;

	for (i = 0; i < p->ssz; i++)
		stream_free(&p->s[i]);
	for (i = 0; i < p->gsz; i++)
		free(p->g[i]);

	free(p->g);
	free(p->s);
	free(p);
}

struct icetools *
icetools_get(const char *server, int verbose)
{
	struct ypdir	  dir;
	struct icetools	 *ic;
	pid_t		  pid;
	int		  fds[2], rc, iodone;
	size_t		  i;
	ssize_t		  ssz;
	char		  buf[BUFSIZ];
	struct pollfd	  fd;
	XML_Parser	  xp;
	void		(*sig)(int);

	error = 0;

	if (NULL == server)
		server = SERVER;

	/*
	 * We don't want to be notified about our child exiting, only
	 * the data that it sends.
	 * To wit, the polL(2) interface will tell us if the data stream
	 * has been terminated.
	 * No need for fancy handling.
	 */
	if (SIG_ERR == (sig = signal(SIGCHLD, SIG_IGN))) {
		perror(NULL);
		return(NULL);
	}

	/*
	 * Create the pipe used to funnel data between the child ftp(1)
	 * process and our XML parser.
	 * Then fork(2) said processes.
	 */

	if (-1 == pipe(fds)) {
		perror(NULL);
		return(NULL);
	} else if (-1 == (pid = fork())) {
		perror(NULL);
		close(fds[0]);
		close(fds[1]);
		return(NULL);
	}

	/*
	 * The child simply connects to the server and spits the data
	 * to stdout, which is connected to the write end of our pipe.
	 */

	if (0 == pid) {
		close(fds[0]);
		dup2(fds[1], STDOUT_FILENO);
		execlp("ftp", "ftp", verbose ? "-v" : "-V",
				"-o", "-", server, NULL);
		perror("ftp");
		exit(EXIT_FAILURE);
	}

	close(fds[1]);

	/*
	 * On the read end, first create an XML parser instance.
	 * Then sit on poll(2), waiting for data from the ftp(1)
	 * subprocess.
	 * When data arrives, push it directly into the XML parser for
	 * the parse itself.
	 */

	memset(&dir, 0, sizeof(struct ypdir));
	dir.state = STATE__MAX;

	if (NULL == (xp = XML_ParserCreate(NULL))) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}

	XML_SetCharacterDataHandler(xp, xml_char);
	XML_SetElementHandler(xp, xml_elem_start, xml_elem_end);
	XML_SetUserData(xp, &dir);

	iodone = 0;

	fd.fd = fds[0];
	fd.events = POLLIN;

	while ((rc = poll(&fd, 1, INFTIM)) > 0) {
		/*
		 * Note that the child process can close (sending us a
		 * HUP) long before we finish consuming data.
		 * Thus, ignore the HUP completely and stop only when
		 * errors occur ("soft", when the parser fails; or
		 * "hard", when a system call fails).
		 */

		if (fd.revents & (POLLERR | POLLNVAL)) {
			fprintf(stderr, "%s: Polling error\n", server);
			break;
		} else if ( ! (POLLIN & fd.revents))
			continue;

		if ((ssz = read(fds[0], buf, sizeof(buf))) < 0) {
			fprintf(stderr, "%s: Read error\n", server);
			break;
		}

		/*
		 * Mark that we've read the entire stream.
		 * We don't really care about errors after that.
		 */

		iodone = 0 == ssz;
		if (0 == XML_Parse(xp, buf, ssz, iodone)) {
			fprintf(stderr, "%s: Parse error\n", server);
			break;
		} else if (error) {
			fprintf(stderr, "%s: XML error\n", server);
			break;
		} else if (iodone)
			break;
	}

	/*
	 * If we've done everything right, put parsed data into our
	 * output structure.
	 * If not, keep it NULL.
	 * Free all ouf our temporary memory.
	 */

	if (rc < 0)
		perror(NULL);

	if (rc >= 0 && iodone && 0 == error) {
		ic = malloc(sizeof(struct icetools));
		if (NULL == ic) {
			perror(NULL);
			exit(EXIT_FAILURE);
		}
		ic->s = dir.s;
		ic->ssz = dir.ssz;
		ic->g = dir.g;
		ic->gsz = dir.gsz;
		ic->gcount = dir.gcount;
	} else {
		ic = NULL;
		for (i = 0; i < dir.ssz; i++)
			stream_free(&dir.s[i]);
		free(dir.s);
		for (i = 0; i < dir.gsz; i++)
			free(dir.g[i]);
		free(dir.g);
	}

	for (i = 0; i < dir.ssz; i++)
		entry_free(&dir.e[i]);

	free(dir.e);
	XML_ParserFree(xp);

	/* Restore signal handler. */

	signal(SIGCHLD, sig);
	return(ic);
}

/*
 * If we're in a known state, append the character data to our current
 * parse handle.
 * (Note: this handler might only pass bits of data at a time.)
 */
static void
xml_char(void *arg, const XML_Char *p, int sz)
{
	struct ypdir	 *dir = arg;
	char		**cp;
	size_t		 *lenp;

	if (error || 0 == sz)
		return;

	cp = NULL;
	lenp = NULL;

	switch (dir->state) {
	case (STATE_NAME):
		cp = &dir->s[dir->ssz - 1].name;
		lenp = &dir->e[dir->ssz - 1].namesz;
		break;
	case (STATE_URL):
		cp = &dir->e[dir->ssz - 1].url;
		lenp = &dir->e[dir->ssz - 1].urlsz;
		break;
	case (STATE_GENRE):
		cp = &dir->e[dir->ssz - 1].genre;
		lenp = &dir->e[dir->ssz - 1].genresz;
		break;
	case (STATE_SONG):
		cp = &dir->e[dir->ssz - 1].song;
		lenp = &dir->e[dir->ssz - 1].songsz;
		break;
	default:
		return;
	}

	*cp = realloc(*cp, *lenp + sz + 1);
	memcpy(*cp + *lenp, p, sz);
	(*cp)[*lenp + sz] = '\0';
	*lenp += sz;
}

/*
 * Begin handling an element.
 * If we recognise the element, first check or state sanity, then shunt
 * into the sub-handler if needed.
 */
static void
xml_elem_start(void *arg, const XML_Char *name, const XML_Char **atts)
{
	size_t		 i;

	if (error)
		return;

	for (i = 0; i < STATE__MAX; i++)  {
		if (strcmp(name, states[i].name))
			continue;
		if (state(arg, states[i].enter, i))
			if (NULL != states[i].start)
				(*states[i].start)(arg);
		return;
	}
}

/*
 * See xml_elem_start().
 */
static void
xml_elem_end(void *arg, const XML_Char *name)
{
	size_t		 i;

	if (error)
		return;

	for (i = 0; i < STATE__MAX; i++)  {
		if (strcmp(name, states[i].name)) 
			continue;
		if (state(arg, i, states[i].enter))
			if (NULL != states[i].end)
				(*states[i].end)(arg);
		return;
	}
}

/*
 * After we've parsed a genre string, we want to look up the
 * space-separated genres passed to us.
 * Look up each parsed genre in the genre table, adding a new element if
 * none is found.
 */
static void
genre(struct ypdir *dir)
{
	char		*start, *end;
	size_t		 i, j;
	struct stream	*e;

	e = &dir->s[dir->ssz - 1];
	start = dir->e[dir->ssz - 1].genre;

	j = 0;
	while (NULL != start && e->gidsz < 32) {
		while (isspace((unsigned char)*start))
			start++;
		end = start;
		while ('\0' != *end && ! isspace((unsigned char)*end))
			end++;
		if (end == start)
			return;
		if ('\0' != *end)
			*end++ = '\0';
		for (i = 0; i < dir->gsz; i++)
			if (0 == strcasecmp(dir->g[i], start))
				break;
		if (i == dir->gsz) {
			dir->g = realloc(dir->g, 
				(i + 1) * sizeof(char *));
			dir->gcount = realloc(dir->gcount, 
				(i + 1) * sizeof(size_t));
			if (NULL == dir->g || NULL == dir->gcount) {
				perror(NULL);
				exit(EXIT_FAILURE);
			}
			dir->gcount[i] = 0;
			dir->g[i] = strdup(start);
			if (NULL == dir->g[i]) {
				perror(NULL);
				exit(EXIT_FAILURE);
			}
			dir->gsz++;
		}

		dir->gcount[i]++;
		e->gids[e->gidsz++] = i;
		start = '\0' == *end ? NULL : end;
	}
}

/*
 * Merge duplicate entities.  Do this after processing.
 * A "duplicate" is one where the name and all genres is the same.
 * If the entity is a duplicate, append its URL to the existing entity's
 * list of URLs and free memory.
 */
static void
dupes(struct ypdir *dir)
{
	size_t		 i, j, len;

	len = dir->ssz - 1;

	for (i = 0; i < len; i++) {
		if (strcmp(dir->s[i].name, dir->s[len].name))
			continue;
		if (dir->s[i].gidsz != dir->s[len].gidsz)
			continue;
		for (j = 0; j < dir->s[i].gidsz; j++)
			if (dir->s[i].gids[j] != dir->s[len].gids[j])
				break;
		if (j < dir->s[i].gidsz)
			continue;

		if (dir->s[i].urlsz < 32) {
			dir->s[i].urls[dir->s[i].urlsz] = 
				dir->e[len].url;
			dir->s[i].songs[dir->s[i].urlsz] = 
				dir->e[len].song;
			dir->e[len].url = dir->e[len].song = NULL;
			dir->s[i].urlsz++;
		}

		for (j = 0; j < dir->s[len].gidsz; j++)
			dir->gcount[dir->s[len].gids[j]]--;

		stream_free(&dir->s[len]);
		entry_free(&dir->e[len]);
		dir->ssz--;
		return;
	}

	dir->s[len].urls[0] = dir->e[len].url;
	dir->s[len].songs[0] = dir->e[len].song;
	dir->s[len].urlsz = 1;
	dir->e[len].url = dir->e[len].song = NULL;
}

/*
 * Construct a new stream entry.
 */
static void
entry(struct ypdir *dir)
{

	if (dir->ssz + 1 > dir->max) {
		dir->max += 4096;
		dir->e = realloc(dir->e, 
			dir->max * sizeof(struct entry));
		dir->s = realloc(dir->s, 
			dir->max * sizeof(struct stream));
		if (NULL == dir->e || NULL == dir->s) {
			perror(NULL);
			exit(EXIT_FAILURE);
		}
	}

	memset(&dir->e[dir->ssz], 0, sizeof(struct entry));
	memset(&dir->s[dir->ssz], 0, sizeof(struct stream));
	dir->ssz++;
}

/*
 * Check our parse state.
 * This ensures that the document is well-formed.
 */
static int
state(struct ypdir *dir, enum state ostate, enum state nstate)
{

	if (ostate != dir->state) {
		error = 1;
		return(0);
	}

	dir->state = nstate;
	return(1);
}

static void
stream_free(struct stream *p)
{
	size_t		 i;

	for (i = 0; i < p->urlsz; i++) {
		free(p->urls[i]);
		free(p->songs[i]);
	}

	free(p->name);
}

static void
entry_free(struct entry *p)
{

	free(p->genre);
	free(p->url);
	free(p->song);
}
